/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.Node;
import java.awt.Color;
import java.awt.Graphics;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;

/**
 *
 * @author Jordan
 */
public class SteinerTree {

    private StNode root_;
    static private double R_;

    public SteinerTree(Node root, ArrayList<Node> destinations, double unitDistance) {
        root_ = new StNode(root);
        R_ = unitDistance;

        // Generate the steiner tree
        // Step 1 - Create steiner nodes for each destination
        ArrayList<StNode> dests = new ArrayList<>();
        for (Node d : destinations) {
            dests.add(new StNode(d));
        }

        // Step 2 - Compute steiner points for each destination pair with the root
        ArrayList<SteinerTuple> pairs = new ArrayList<>();
        for (StNode u : dests) {
            for (StNode v : dests) {
                if (u.destination().id() <= v.destination().id()) {
                    pairs.add(new SteinerTuple(root_, u, v));
                }
            }
        }

        // Step 3 - Sort all steiner tuples (root, left, right, steiner) by their reduction ratios.
        try {
            Collections.sort(pairs);
        } catch (IllegalArgumentException e) {
            for (SteinerTuple pair : pairs) {
                System.out.println(pair);
            }
        }

        boolean sort = false;   // flag to indicate sort required for new nodes at end of round
        boolean clean = false;  // flag to indicate removal of inactive nodes at end of round
        StNode s = root_;       // 's' is current node being constructed upon - starts from root

        // Step 4 - Construct the tree iteratively
        while (!pairs.isEmpty()) {
            SteinerTuple best = pairs.get(0);
            StNode u = best.u();
            StNode v = best.v();
            StNode t = best.t();

            // Choose conditional prioritized build action:
            // 1) : IF u == v then s adds u|v as child
            if (u == v) {
                s.addChild(u);
                u.deactivate();
                dests.remove(u);
                dests.remove(v);
                clean = true;
            } // 2) : IF t collocated with s then s adds u and v as children
            else if (t.collocated(s)) {
                s.addChild(u);
                s.addChild(v);
                u.deactivate();
                v.deactivate();
                dests.remove(u);
                dests.remove(v);
                clean = true;
            } // 3) : IF t collocated with u then u adds v as child
            else if (t.collocated(u)) {
                u.addChild(v);
                v.deactivate();
                dests.remove(v);
                clean = true;
            } // 4) : IF t collocated with v then v adds u as child
            else if (t.collocated(v)) {
                v.addChild(u);
                u.deactivate();
                dests.remove(u);
                clean = true;
            } // 5) : IF d(s, u) < R AND d(s, v) < R then remove (u, v)
            else if (s.distance(u) < R_ && s.distance(v) < R_) {
                pairs.remove(best);
            } // 6) : IF only d(s, u) < R :
            else if (s.distance(u) < R_) {
                // IF R + d(t, u) + d(t, v) > d(s, u) + d(s, v) THEN remove (u, v)
                if (R_ + t.distance(u) + t.distance(v) > s.distance(u) + s.distance(v)) {
                    pairs.remove(best);
                } // ELSE u adds v as child
                else {
                    u.addChild(v);
                    v.deactivate();
                    dests.remove(v);
                    clean = true;
                }
            } // 7) : IF only d(s, v) < R :
            else if (s.distance(v) < R_) {
                // IF R + d(t, u) + d(t, v) > d(s, u) + d(s, v) THEN remove (u, v)
                if (R_ + t.distance(u) + t.distance(v) > s.distance(u) + s.distance(v)) {
                    pairs.remove(best);
                } // ELSE v adds u as child
                else {
                    v.addChild(u);
                    u.deactivate();
                    dests.remove(u);
                    clean = true;
                }
            } // 8) : IF d(s, t) < R AND R + d(t, u) + d(t, v) > d(s, u) + d(s, v) THEN s adds u and v as children
            else if (s.distance(t) < R_ && R_ + t.distance(u) + t.distance(v) > s.distance(u) + s.distance(v)) {
                s.addChild(u);
                s.addChild(v);
                u.deactivate();
                v.deactivate();
                dests.remove(u);
                dests.remove(v);
                clean = true;
            } // 9) : Default action - create a new virtual destination at the steiner point. Note: virtual dests
            //      are never 'visited'.
            else {
                t.addChild(u);
                t.addChild(v);
                dests.add(t);

                for (StNode d : dests) {
                    pairs.add(new SteinerTuple(root_, t, d));
                }

                u.deactivate();
                v.deactivate();
                dests.remove(u);
                dests.remove(v);

                // Require sort for newly added (t, d) pairs and cleaning for deactivating destinations.
                clean = true;
                sort = true;
            }

            // End of round
            // - Remove any pairs with deactivated destinations
            if (clean) {
                Iterator<SteinerTuple> pairIt = pairs.iterator();

                while (pairIt.hasNext()) {
                    SteinerTuple pair = pairIt.next();
                    if (!pair.u().active() || !pair.v().active()) {
                        pairIt.remove();
                    }
                }

                // Clear flag for next round
                clean = false;
            }

            // - Re-sort the pair collection for newly added pairs
            if (sort) {
                try {
                    Collections.sort(pairs);
                } catch (IllegalArgumentException e) {
                    for (SteinerTuple pair : pairs) {
                        System.out.println(pair);
                    }
                }

                // Clear flag for next round
                sort = false;
            }
        }
    }
    
    public SteinerTree(StNode root, double unitDistance){
        root_ = root.getClone();
        R_ = unitDistance;
    }

    public StNode root() {
        return root_;
    }
    
    public double RR(){
        return R_;
    }

    public void paint(Graphics g) {
        root_.destination().paint(g, Color.red);

        for (StNode child : root_.children()) {
            child.paint(g);
        }
    }    
}
